// kajtest.cc -- main program for my subscription to the OpenGL StackHack
// http://www.stacken.kth.se/stackhack/20051110

// "THE BEER-WARE LICENSE" (Revision 42):
// <kaj@stacken.kth.se> wrote this file. As long as you retain this notice you
// can do whatever you want with this stuff. If we meet some day, and you
// think this stuff is worth it, you can buy me a beer in return. Rasmus Kaj.

#include <iostream>
#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <GL/glfw.h>
#include <math.h>

#define WIDTH 640
#define HEIGHT 480

// Load a wavefront OBJ file from the specified filename.
// implemented in loadobj.cc
void loadObj(const std::string& filename);

// Check the gl status, if there is an error, log the status and the "when"
// message to standard error.
void checkGl(const std::string& when) {
  GLenum status;
  while((status = glGetError()) != GL_NO_ERROR) {
    std::cerr << "GL error " << status << " at " << when << std::endl;
  }
}

// Callback for window resize
static void GLFWCALL size_callback(int width, int height) {
  glViewport(0, 0, width, height);
}

// Draw a recursively defined tree
// part is the id of a calllist for a "trunc part"
// t is a "time" variable to make the tree move
// depth is how deep to recurse.
void drawTree(GLuint part, double t, int depth) {
  int num = 6 - depth;
  glCallList(part);
  glTranslatef(0.f, 0.f, 0.8f);
  if(depth > 1) for(int i = 0; i < num; ++i) {
    glPushMatrix();
    glRotatef(t * 30.f + 360.f*i/num, 0.f, 0.f, 1.f);
    
    glRotatef(30 + 20*sin(1.5*t + 2*M_PI*i/num), 1.f, 0.f, 0.f);
    glScalef(0.9f, 0.5f, 0.8f);
    drawTree(part, t, depth -1);
    glPopMatrix();
  }
}

// Call this method once per frame to keep track of the frame rate and report
// to standard output.
void fps() {
  static double prevtime = glfwGetTime();
  static int n = 0;
  const int frames = 50;	// number of frame to measure over
  if(++n >= frames) {
    
    double time = glfwGetTime();
    std::cout << "Rate is " << frames / (time - prevtime) << " FPS."
	      << std::endl;
    
    prevtime = time;
    n = 0;
  }
}

// Create a tree trunk as a specific call-list id (must allready be declared
// by genLists.
void createTrunk(GLuint list) {
  glNewList(list, GL_COMPILE); 
  const int num_colors = 3;
  GLfloat color[num_colors][3] = {
    { 0.7f, 0.8f, 0.3f },
    { 0.9f, 0.7f, 0.4f },
    { 0.6f, 0.3f, 0.1f },
  };
  glLightfv(GL_LIGHT0, GL_SPECULAR, color[1]);
  glMaterialfv(GL_FRONT, GL_DIFFUSE|GL_AMBIENT, color[0]);
  
  glBegin(GL_TRIANGLE_FAN);
  const int num_sides = 7;
  glNormal3f(0.f, 0.f, -1.0f);
  glVertex3f(0.f, 0.f, 0.f);
  for(int i = 0; i <= num_sides; ++i) {
    const float fi = 2*M_PI*i/num_sides;
    const float r1 = 0.8f, r2 = 0.1f; 
    glMaterialfv(GL_FRONT, GL_DIFFUSE|GL_AMBIENT, color[i%num_colors]); 
    glNormal3f(r1*sin(fi), r1*cos(fi), -0.6f);
    glVertex3f(r2*sin(fi), r2*cos(fi), 0.1f);
  }
  glEnd();
  
  glBegin(GL_TRIANGLE_STRIP);
  for(int i = 0; i <= num_sides; ++i) {
    const float fi = 2*M_PI*i/num_sides;
    const float fi2 = 2*M_PI*(i+.5)/num_sides;
    const float r1 = 0.8f, r2 = 0.1f, r3 = 0.2f;
    glMaterialfv(GL_FRONT, GL_DIFFUSE|GL_AMBIENT, color[i%num_colors]);
    glNormal3f(r1*sin(fi),  r1*cos(fi), -0.6f);
    glVertex3f(r2*sin(fi),  r2*cos(fi),  0.1f);
    glNormal3f(r1*sin(fi2), r1*cos(fi2), 0.6f);
    glVertex3f(r3*sin(fi2), r3*cos(fi2), 1.f);
  }
  glEnd();
  
  glBegin(GL_TRIANGLE_FAN);
  glNormal3f(0.f, 0.f, 1.0f);
  glVertex3f(0.f, 0.f, 1.1f);
  for(int i = num_sides; i >= 0; --i) {
    const float fi = 2*M_PI*(i+.5)/num_sides;
    glMaterialfv(GL_FRONT, GL_DIFFUSE|GL_AMBIENT, color[i%num_colors]);
    glNormal3f(0.8*sin(fi), 0.8*cos(fi), 0.6f);
    glVertex3f(0.2f*sin(fi), 0.2f*cos(fi), 1.0f);
  }
  glEnd();
  glEndList();
}

void drawNet(int lnum, float lsep) {
  glBegin(GL_LINES);
  glColor3f(0.5f, 1.0f, 0.5f);
  for(int i = -lnum; i < lnum; ++i) {
    glVertex3f(-lnum*lsep, i*lsep, 0.0f);
    glVertex3f( lnum*lsep, i*lsep, 0.0f);
  }
  for(int i = -lnum; i < lnum; ++i) {
    glVertex3f(i*lsep, -lnum*lsep, 0.0f);
    glVertex3f(i*lsep,  lnum*lsep, 0.0f);
  }
  glEnd();
}

int main(int argc, char *argv[])
{
  GLuint list;
  //int x, y, z;
  GLfloat color[4] = { 0.f, 0.f, 0.f, 0.f };
  GLfloat lightcolor[4] = { 0.2f, 0.2f, 0.2f, 0.2f };
  GLfloat position[4] = { 2.f, -5.f, 2.f, 0.f };
  if (!glfwInit())
    return 1;

  if (glfwOpenWindow(WIDTH, HEIGHT, 8, 8, 8, 0, 16, 0, GLFW_WINDOW))
  {
    glfwSetWindowTitle(argv[0]);
    glfwSetWindowSizeCallback(size_callback);
    
    glViewport(0, 0, WIDTH, HEIGHT);
    glClearColor(color[0], color[1], color[2], color[3]);
    glEnable(GL_CULL_FACE);
    
    
    glMaterialfv(GL_FRONT, GL_DIFFUSE, (GLfloat*) &color);
    
    glEnable(GL_DEPTH_TEST);
//      glEnable(GL_BLEND);
//      glBlendFunc(GL_ONE, GL_ONE);
    
    glEnable(GL_FOG);
    glFogi(GL_FOG_MODE, GL_EXP);
    glFogf(GL_FOG_DENSITY, 0.03f);
//     glFogf(GL_FOG_START, 3.f);
//     glFogf(GL_FOG_END, 10.f);
    glFogfv(GL_FOG_COLOR, (GLfloat*) &color);
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(60.f, (GLdouble) WIDTH / (GLdouble) HEIGHT, 0.1f, 100.f);
    glMatrixMode(GL_MODELVIEW);
    checkGl("before lists");
    
    list = glGenLists(3);
    createTrunk(list);
    checkGl("after tree list");

    glNewList(list+1, GL_COMPILE);
    drawNet(20, 0.9);
    glEndList();
    checkGl("after net list");
    
    glNewList(list+2, GL_COMPILE);
    loadObj("lightcycle-high.obj");
    glEndList();
    checkGl("after loaded list");

    while (glfwGetWindowParam(GLFW_OPENED))
    {
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      glEnable(GL_LIGHT0);
      //glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);
      glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SINGLE_COLOR);
      glLightfv(GL_LIGHT0, GL_POSITION, position);
      glLightfv(GL_LIGHT0, GL_AMBIENT, lightcolor);
      glLightfv(GL_LIGHT0, GL_DIFFUSE, lightcolor);
      glLightfv(GL_LIGHT0, GL_SPECULAR, lightcolor);
      double t = glfwGetTime();
      
      glLoadIdentity();
      glTranslatef(0.f, 0.f, -8.f);
      glRotatef(-75.f, 1.0f, 0.f, 0.f);
      float scale = 1.3f;
      glScalef(scale, scale, scale);
      checkGl("loop init");


      glRotatef(t * 11.f, 0.f, 0.f, 1.f);
      
      glDisable(GL_LIGHTING);
      glCallList(list+1);
      checkGl("loop net");

      glEnable(GL_LIGHTING);
#if 1
      glPushMatrix();
      glTranslatef(-2*.9f, -2*.9f, -.1f);
      glRotatef(10.f, 1.0f, 0.0f, 0.0f);
      drawTree(list, 1.5*t, 3);
      glPopMatrix();
      
      glPushMatrix();
      glTranslatef( 2*.9f, -2*.9f, -.1f);
      glRotatef(10.f, 1.0f, 0.0f, 0.0f);
      drawTree(list, t, 4);
      glPopMatrix();
      
      glPushMatrix();
      glTranslatef(-2*.9f,  2*.9f, -.1f);
      glRotatef(10.f, 1.0f, 0.0f, 0.0f);
      drawTree(list, 0.6*t, 4);
      glPopMatrix();
      
      glPushMatrix();
      glTranslatef( 2*.9f,  2*.9f, -.1f);
      glRotatef(10.f, 1.0f, 0.0f, 0.0f);
      drawTree(list, 0.3*t, 3);
      glPopMatrix();
      checkGl("loop trees");
#endif

      for(int i = 0; i < 4; ++i) {
	float cyclepos = 25*0.9*((fmod(t, 3.0) - 1.5f)/1.5f + (i/4.f-.5f));
	
	glPushMatrix();
	glRotatef(i*90.f, 0.f, 0.f, 1.f);
	glTranslatef(-0.9f, -cyclepos, 0.f);
	glScalef(0.3f, 0.3f, 0.3f);
	glTranslatef(0.f, 0.f, 2.f);
	glColor3f(0.f, 0.0f, 1.f);
	glCallList(list+2);
	glPopMatrix();
	checkGl("loop cycle i");
      }
      
      glfwSwapBuffers();
      checkGl("loop swapbuffers");
      
      fps();
      
    }

    glDeleteLists(list, 1);
  }

  glfwTerminate();
  return 0;
}

