// loadobj.cc -- Load a Wavefront OBJ, part of my subscription to the OpenGL
// StackHack http://www.stacken.kth.se/stackhack/20051110

// "THE BEER-WARE LICENSE" (Revision 42):
// <kaj@stacken.kth.se> wrote this file. As long as you retain this notice you
// can do whatever you want with this stuff. If we meet some day, and you
// think this stuff is worth it, you can buy me a beer in return. Rasmus Kaj.

#include <GL/glfw.h>
#include <string>
#include <iostream>
#include <vector>
#include <fstream>
#include <sstream>
#include <map>

class Vector3 {
public:
  Vector3(float xx, float yy, float zz)
    {
      v[0] = xx; v[1] = yy; v[2] = zz;
    }
  
  GLfloat v[3];
};

typedef std::vector<Vector3> VectorVec;


class Material {
  typedef std::vector<float> Param;
  typedef std::vector<std::pair<GLenum, Param> > ParamSet;

public:  
  void setParam(GLenum pname, std::istream& values) {
    std::vector<float> val;
    float v;
    while(values >> v) val.push_back(v);
    params.push_back(std::make_pair(pname, val));
  }
  
  void apply() {
    for(ParamSet::const_iterator i = params.begin(); i != params.end(); ++i) {
      const Param& val = i->second;
      if(val.size() == 1) {
	glMaterialf(GL_FRONT, i->first, val[0]);
      } else {
	GLfloat *f = new GLfloat(val.size());
	for(unsigned int j = 0; j < val.size(); ++j)
	  f[j] = val[j];
	glMaterialfv(GL_FRONT, i->first, f);
	delete[] f;
      }
    }
  }
  
  ParamSet params;
};

typedef std::map<std::string, Material*> MaterialMap;

MaterialMap loadMaterials(const std::string& filename) {
  std::ifstream mtlfil(filename.c_str());
  std::string line;

  MaterialMap result;
  Material* mtl = 0;
  
  while(getline(mtlfil, line)) {
    if(line.length() > 0 && line[0] != '#') {
      // non-empty, non-comment line
      std::istringstream ll(line);
      std::string cmd;
      ll >> cmd;
      if(cmd == "newmtl") {
	std::string name;
	ll >> name;
	mtl = new Material();
	result[name] = mtl;

      } else if(cmd == "Ka") {
	mtl->setParam(GL_AMBIENT, ll);
      } else if(cmd == "Kd") {
	mtl->setParam(GL_DIFFUSE, ll);
      } else if(cmd == "Ks") {
	mtl->setParam(GL_SPECULAR, ll);
      } else if(cmd == "Ns") {
	mtl->setParam(GL_SHININESS, ll); // possibly?
      } else if(cmd == "Ni") {
	// todo mtl->setParam(NI, ll);
      } else if(cmd == "illum") {
	// todo mtl->setParam(ILLUM, ll);
      } else {
	std::cerr << "Unknown material parameter " << cmd << std::endl;
      }
    }
  }
  return result;
}

void loadObj(const std::string& filename) {
  std::ifstream obj(filename.c_str());

  VectorVec vertices, normals;
  vertices.push_back(Vector3(0,0,0));
  normals.push_back(Vector3(0,0,0));
  
  MaterialMap material;
  bool begun = false;
  
  std::string line;
  while(getline(obj, line)) {
    if(line.length() > 0 && line[0] != '#') {
      // non-empty, non-comment line
      std::istringstream ll(line);
      std::string cmd;
      ll >> cmd;
      if(begun && cmd != "f") {
	glEnd();
	begun = false;
      }
      
      if(cmd == "mtllib") {
	std::string filename;
	ll >> filename;
	material = loadMaterials(filename);
	
      } else if(cmd == "v") {	// vertex
	float x, y, z;
	ll >> x >> y >> z;
	vertices.push_back(Vector3(x, y, z));
	
      } else if(cmd == "vn") {	// normal
	float x, y, z;
	ll >> x >> y >> z;
	normals.push_back(Vector3(x, y, z));

      } else if(cmd == "g") {	// What's this? TODO

      } else if(cmd == "usemtl") { // use named material.  TODO
	std::string mtl;
	ll >> mtl;
	Material* m = material[mtl];
	if(m) m->apply();
	else
	  std::cerr << "Unknown material \"" << mtl << "\" for usemtl in "
		    << filename << std::endl;
	
      } else if(cmd == "f") {
	int v1, v2, v3, n1, n2, n3;
	char c[6];
	ll >> v1 >> c[0] >> c[1] >> n1
	   >> v2 >> c[2] >> c[3] >> n2
	   >> v3 >> c[4] >> c[5] >> n3;
	if(!begun) {
	  glBegin(GL_TRIANGLES);
	  begun = true;
	}
	glNormal3fv(normals.at(n1).v);
	glVertex3fv(vertices.at(v1).v);
	glNormal3fv(normals.at(n2).v);
	glVertex3fv(vertices.at(v2).v);
	glNormal3fv(normals.at(n3).v);
	glVertex3fv(vertices.at(v3).v);
//	glEnd();
      } else {
	std::cerr << "Unknown keyword \"" << cmd << "\" in " << filename
		  << std::endl;
      }
    }
  }
  if(begun)
    glEnd();
}
